/************************************************************************/
/*  FbtServer.c                                                         */  
/*  B. Mours, D.Verkindt (LAPP Annecy)                                  */
/************************************************************************/
#include "FdIO.h"
#include "Frv.h"

/*-----------------------------------------------------private definitions---*/
typedef struct FdFbt FdFbt;

struct FdFbt               /* Trend frame builder parameters          */
{
  char   *tag;             /* tag to compute remaining trend data     */
  int     frameL;          /* trend frame length (in seconds)         */
  FrFile *oFile;           /* output trend file pointer               */
  int     nChannels;       /* number of channel in the trend frame    */
  int     nAdc;            /* number of ADC channels                  */
  int     adcSize;         /* size of the input ADC's                 */
  double  procTime;        /* processing time                         */
  FrameH *trend;           /* current trend frame                     */
  FrCList *listTrend;      /* fast access list for the trend channels */
  FdAction*  action;      /* action object which hold this object */
};

void FdFbtProcess(FdFbt *fbt, FrameH *frame);

/*---------------------------------------------------------------------------*/
void  FbtFillAdc(FdFbt *fbt,
		 char  *name,
		 char  *word1,
		 double value,
		 int    index,
		 int    nData,
		 double samplingRate,
		 double slope,
		 double bias,
		 char  *units,
		 int   nBits)
/*---------------------------------------------------------------------------*/
/* This function fill (and create it if needed) an trend adc channel         */
/*---------------------------------------------------------------------------*/
{
  FrameH *trend;
  FrAdcData *adc;
  FrVect  *isThere;
  int  j, k;
  char channelName[512];
 
  trend = fbt->trend;

  /*------------------------ search the adc channel in the trend frame---*/
  if(word1 != NULL) sprintf(channelName, "%s_%s",name, word1);
  else              sprintf(channelName, "%s",   name);
  adc = (FrAdcData *) FrCListFind(fbt->listTrend,channelName);

  /*-------------- the channel is not in the trend frame. Create it ---*/
  if (adc == NULL) {

     adc = FrAdcDataNewF(trend, channelName, NULL, 0, 0, nBits, 
			bias, slope, units, samplingRate, nData);
    if (adc == NULL) {
      CfgMsgAddFatal("FbtBuildTrendFrame: malloc adc failed. "
		     "channel=%s  nData=%d  nBits=%d",
		     channelName, nData, nBits);}
    for(j=0; j<adc->data->nData; j++) {adc->data->dataI[j] = 0;}

    /*----------------------------------------add auxiliary vector ---*/
    isThere = FrVectNew1D("Available_data",  FR_VECT_C,  nData, 
			  adc->data->dx[0], adc->data->unitX[0], 
			  "1 if data is available");
    if(isThere == NULL) {
      CfgMsgAddFatal("FbtBuildTrendFrame: malloc failed"
		     " channel=%s  nData=%d", channelName, nData);}
    for(k=0; k<isThere->nData; k++) {isThere->data[k] = 0.;}
    adc->data->next = isThere;}

  /*-----------------------------adc channel is there, fill data ---*/
  if (index >= adc->data->nData) {
    CfgMsgAddError("FbtBuildTrendFrame. Vector overflow at %d index=%d "
		   "samplingRate=%g  channel=%s serData:%s",
		   trend->GTimeS, index, samplingRate, channelName,
		   name);
    CfgMsgAddError("  nData=%"FRLLD" nbits=%d",adc->data->nData, nBits);
    index =  adc->data->nData - 1;}

  if(nBits > 0) adc->data->dataUI[index] = value;
  else          adc->data->dataF[index]  = value;
  adc->data->next->data[index] ++;

  return;
} 
/*------------------------------------------------------ FbtBuildTrendData --*/
void FbtBuildTrendData (FdFbt *fbt, FrameH *frame)
/*---------------------------------------------------------------------------*/
/* This function build and add trend data for each taged ADC channel.        */
/*---------------------------------------------------------------------------*/
{
  FrAdcData *adc;
  FrVect *data;
  FrProcData *proc;
  double min, max, mean, rms, samplingRate, trendRate, m, slope, bias;
  int index, nData;

  index = (frame->GTimeS % fbt->frameL)/frame->dt;
  nData =  fbt->frameL/frame->dt;
  trendRate = 1./frame->dt;

  fbt->nChannels = 0;

  /*---------------------------------- build the trend data for the adc's---*/
  adc = NULL;
  if (frame->rawData != NULL) adc = frame->rawData->firstAdc;

  for (; adc != NULL; adc = adc->next) {
    data = adc->data;
    if(data->nDim != 1) continue;
    if(FrVectIsValid(data) != 0) continue;
    fbt->nChannels++;

    if(data->nData == 1) {
      samplingRate = 1.;
      if(data->dx[0] > 1) samplingRate = 1./data->dx[0];
      FbtFillAdc(fbt,adc->name,NULL, FrVectGetValueI(data,0),
                 index*frame->dt*samplingRate, fbt->frameL*samplingRate, 
                 samplingRate, adc->slope,
                 adc->bias, adc->units, -32);
      continue;}

    FrVectMinMax(data, &min, &max);
    FrvRms(data, &mean, &rms);
    if(adc->slope < 0) {
      slope = -adc->slope;
      bias  = -adc->bias;
      m = max;
      max = -min; 
      min = -m;
      mean = -mean;}
    else {    
      slope = adc->slope;
      bias  = adc->bias;}
  
    FbtFillAdc(fbt, adc->name,"min", min,   index, nData, trendRate,
	       slope, bias, adc->units, -32);
    FbtFillAdc(fbt, adc->name,"mean", mean, index, nData, trendRate,
	       slope, bias, adc->units, -32);
    FbtFillAdc(fbt, adc->name,"max", max,   index, nData, trendRate,
	       slope, bias, adc->units, -32);
    FbtFillAdc(fbt, adc->name,"rms", rms,   index, nData, trendRate,
	       slope, bias, adc->units, -32);}

  /*-------------------------------------- build trend data for FrProcData---*/
  for (proc = frame->procData; proc != NULL; proc = proc->next) {
    data = proc->data;
    if(data->nDim != 1) continue;
    if(FrVectIsValid(data) != 0) continue;
    fbt->nChannels++;

    if(data->nData == 1) {/*--- one 1Hz signal: copy only the value--*/
      samplingRate = 1.;
      if(data->dx[0] > 1) samplingRate = 1./data->dx[0];
      FbtFillAdc(fbt, proc->name,NULL, FrVectGetValueI(data,0), 
		 index*frame->dt*samplingRate, fbt->frameL*samplingRate, 
                 samplingRate, 1., 0., data->unitY, -32);
      continue;}

    FrVectMinMax(data, &min, &max);
    FrvRms(data, &mean, &rms);

    FbtFillAdc(fbt, proc->name, "min",  min, index, nData, trendRate,
	       1., 0., data->unitY, -32);
    FbtFillAdc(fbt, proc->name,"mean", mean, index, nData, trendRate,
	       1., 0., data->unitY, -32);
    FbtFillAdc(fbt, proc->name, "max",  max, index, nData, trendRate,
	       1., 0., data->unitY, -32);
    FbtFillAdc(fbt,  proc->name, "rms", rms, index, nData, trendRate,
	       1., 0., data->unitY, -32);}

  return;
}
/*---------------------------------------------------------------------------*/
void  FbtBuildTrendFrame(FdFbt *fbt, FrameH *frame)
/*---------------------------------------------------------------------------*/
/* This function convert all FrSerData to a trend frame.                     */
/*---------------------------------------------------------------------------*/
{
  FrDetector *det, **lastDet;
  FrSerData *sms;
  int  nData, i, index, period, nRead;
  unsigned int valueI;
  char *data, word1[256], word2[256], word3[256];
  char channelName[512], units[256];
  float bias, slope;
  double value;
 
  if (frame == NULL) return;

  index = frame->GTimeS % fbt->frameL; /*--compute index in the trend frame--*/

  /*---------------------------------- create a new trend frame if needed ---*/
  if (fbt->trend == NULL)  {
    fbt->trend = FrameHNew("Trend_frame");
    if (fbt->trend == NULL) CfgMsgAddFatal("could not create trend frame");
    fbt->trend->run       = 0;
    fbt->trend->frame     = frame->GTimeS / fbt->frameL;
    fbt->trend->GTimeS    = frame->GTimeS - index;
    fbt->trend->GTimeN    = 0;
    fbt->trend->ULeapS    = frame->ULeapS;
    fbt->trend->dt        = fbt->frameL;

    lastDet = &(fbt->trend->detectProc);
    for(det = frame->detectProc; det != NULL; det = det->next) {
      *lastDet = FrDetectorNew(det->name);
      lastDet = &((*lastDet)->next);
      fbt->trend->detectProc->prefix[0]    = det->prefix[0];
      fbt->trend->detectProc->prefix[1]    = det->prefix[1];
      fbt->trend->detectProc->longitude    = det->longitude;
      fbt->trend->detectProc->latitude     = det->latitude;
      fbt->trend->detectProc->elevation    = det->elevation;
      fbt->trend->detectProc->armXazimuth  = det->armXazimuth;
      fbt->trend->detectProc->armYazimuth  = det->armYazimuth;
      fbt->trend->detectProc->armXaltitude = det->armXaltitude;
      fbt->trend->detectProc->armYaltitude = det->armYaltitude;
      fbt->trend->detectProc->armXmidpoint = det->armXmidpoint;
      fbt->trend->detectProc->armYmidpoint = det->armYmidpoint;
      fbt->trend->detectProc->localTime    = det->localTime;}
  }

  /*---------- build a fast access list for the channel in the trend frame---*/
  fbt->listTrend = FrCListBldAdc(fbt->trend);

  /*------------------------convert sms data to adc vectors ---*/
  if (frame->rawData == NULL) return;
  for(sms = frame->rawData->firstSer; sms != NULL; sms = sms->next) {
  
    if (sms->sampleRate <= 0) sms->sampleRate = 1.;
    period = (0.5 + 1. / sms->sampleRate);

    if(fbt->frameL % period != 0) {
      CfgMsgAddError("At %d: the trend frame length (%d) is not "
                     "a multiple of the period (%d) of %s",
                     frame->GTimeS,fbt->frameL, period, sms->name);
      continue;}

    i      = index / period;
    nData  = fbt->frameL / period;
    if(nData < 1) nData = 1;

    /*------------------------------------- decode each individual blocks ---*/
    strcpy(units,"Counts");
    bias  = 0.;
    slope = 1.;
    data = sms->data;
    while ((nRead = sscanf(data,"%255s %255s %255s",word1,word2,word3)) >= 2) {
  
      if(strlen(word1) > 200) {
	CfgMsgAddError("FbtBuildTrendFrame: %d too long word for %s: %s",
		       frame->GTimeS, sms->name, word1);
	break;}

      /*------------------------------------ save calibration info if any ---*/
      if     (strcmp(word1,"bias")  == 0)  {sscanf (word2,  "%e", &bias);} 
      else if(strcmp(word1,"slope") == 0)  {sscanf (word2,  "%e", &slope);} 
      else if(strcmp(word1,"units") == 0)  {strcpy(units,word2);}

      /*------------------------ search the adc channel in the trend frame---*/
      else {
        sprintf(channelName,"%s_%s",sms->name, word1);

        if (word2[0] == 'x') {
	  sscanf (word2+1,"%x", &valueI);
	  FbtFillAdc(fbt, sms->name, word1, valueI, i, nData, 1./period,
		     slope, bias, units, 32);}
        else if(strncmp(word2,"0x",2) == 0) {
	  sscanf (word2,  "%x", &valueI);
	  FbtFillAdc(fbt, sms->name, word1, valueI, i, nData, 1./period,
		     slope, bias, units, 32);}
        else if(isalpha((int) word2[0]) == 0) {
	  sscanf (word2,  "%le", &value);
	  FbtFillAdc(fbt, sms->name, word1, value, i, nData, 1./period,
		     slope, bias, units, -32);}
        else  {
	  sscanf (word2+1,"%le", &value);
	  FbtFillAdc(fbt, sms->name, word1, value, i, nData, 1./period,
		     slope, bias, units, -32);}}

      if (nRead != 3) break;
      data = strstr(data+strlen(word1)+strlen(word2), word3);}
  }

  return;
}
/*---------------------------------------------------------------------------*/
void FbtCleanAuxVector(FrameH *frame)
/*---------------------------------------------------------------------------*/
/* This function remove the auxiliary vectors if all the bins are filled     */
/*---------------------------------------------------------------------------*/
{
  FrAdcData *adc;
  FrVect  *isThere;
  int i;

  if (frame == NULL) return;
  if (frame->rawData == NULL) return;

  for(adc = frame->rawData->firstAdc; adc != NULL; adc = adc->next)  {
    isThere = adc->data->next;
    if(isThere == NULL) continue;

    for(i=0; i<isThere->nData; i++) {
      if(isThere->data[i] == 0) break;}
     
    if(i == isThere->nData) {
      FrVectFree(isThere);
      adc->data->next = NULL;}}

  return;
}
/*---------------------------------------------------------------------------*/
int FdFbtNew(void *arg, CfgDataType_t *data)
/*---------------------------------------------------------------------------*/
{
  FdFbt *fbt;

  /*------------create the object and add it at the end of the linked list---*/
  fbt = (FdFbt*) calloc(1,sizeof(FdFbt));
  if(fbt == NULL) CfgMsgAddFatal("malloc FdFbt failed");

  fbt->action = FdActionNew((FdAction**) arg, 
			    (void*) fbt, FdFbtProcess, "Fbt");

  /*------------------------------------------------------------ fill info---*/
  FrStrCpy(&(fbt->tag),       CfgParseGetNextString(data));
  fbt->frameL          =      CfgParseGetNextDec(data);
 
  fbt->trend  = NULL;
  fbt->oFile  = NULL;

  CfgMsgAddInfo("Add trend frame building; tag=%s trend frame length=%ds", 
		fbt->tag, fbt->frameL);
  if(fbt->frameL < 1) CfgMsgAddFatal("Too small trend frame length");

  return(CFG_OK);
}
/*-----------------------------------------------------------FdFbtParseAdd---*/
void FdFbtParserAdd(FdIO *fdIO)
/*---------------------------------------------------------------------------*/
{
  CfgParseGetFunctionAdd(fdIO->parser, "FDOUT_TREND_FRAME",
			 FdFbtNew, (void *) &(fdIO->actionsOut), 2,
			 CfgString, CfgDec);

  return;
}
/*----------------------------------------------------FdFbtProcessOneFrame---*/
FrameH* FdFbtProcessOneFrame(FdFbt *fbt, FrameH *frame)
/*---------------------------------------------------------------------------*/
{
  struct timeb tm1,tm2;
  FrameH *frameReturned;

  ftime(&tm1);

  /*------------------------ write uncomplete trend frame in case of a gap---*/
  frameReturned = NULL;
  if(fbt->trend != NULL) {
    if(fbt->trend->GTimeS + fbt->trend->dt <= frame->GTimeS) {
      FbtCleanAuxVector(fbt->trend);
      frameReturned = fbt->trend;
      fbt->trend = NULL;}}

  /*---------------------------------------- select the channel to be used---*/
  FrameTag(frame, fbt->tag);

  /*-------------------------------------Convert ser data to a trend frame---*/
  FbtBuildTrendFrame (fbt, frame);

  /*--------------------------------- compute trend data for ADC and proc ---*/
  FbtBuildTrendData (fbt, frame);

  FrCListFree(fbt->listTrend);
  fbt->listTrend = NULL;
 
  /*---------------------------------------------- output frame if needed ---*/
  if(fbt->trend->GTimeS + fbt->trend->dt == frame->GTimeS + 1 &&
     frameReturned == NULL) {
    FbtCleanAuxVector(fbt->trend);
    frameReturned = fbt->trend;
    fbt->trend = NULL;}
 
  FrameUntag(frame);

  ftime(&tm2);
  fbt->procTime  = (tm2.time + 1.e-3*tm2.millitm) -
    (tm1.time + 1.e-3*tm1.millitm);

  /*------- free the input frame and return the trend frame if available-----*/
  FrameFree(frame);
  return(frameReturned);
}
/*---------------------------------------------------------------------------*/
void FdFbtProcess(FdFbt *fbt,
		  FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FdAction *next;

  if(frame != NULL) {
    frame = FdFbtProcessOneFrame(fbt, frame);
    if(frame == NULL) return;}

  else { /*--------------------------------------call for the termination---*/
    if(fbt->oFile != NULL) {
      FrFileOEnd(fbt->oFile);
      fbt->oFile = NULL;}
    else {
      if(fbt->trend != NULL) {
	FbtCleanAuxVector(fbt->trend);
        next = fbt->action->next;  /*------- to write the last trend frame---*/
        if(next != NULL) next->action(next->data, fbt->trend);}}}

  /*------------------------------------------call next action in the list---*/
  next = fbt->action->next;
  if(next != NULL) next->action(next->data, frame);

  return;
}
