#include "FdIO.h"
#include <errno.h>
#include <sys/stat.h>

/*-------------------------------------------------------private functions---*/

typedef struct FdIFile FdIFile;
typedef struct FdOFile FdOFile;

struct FdIFile     
{
  char*     name;          /* output file name                              */
  int       start;         /* start GPS time; 0=file start; -1= now         */
  int       length;        /* duration to be process or -1 for the full file*/
  double    requestedTime; /* GPS time of the next frame to be read         */
  int       endTime;       /* end of the frame segment to be read           */
  FrFile*   iFile;         /* output file pointer                           */
  FdAction* action;        /* this is the action structure                  */
};

struct FdOFile     
{
  char*     name;          /* output file name                              */
  int       fLength;       /* number of frame per output file               */
  int       dirPeriod;     /* period cover by a directory or 0 if single dir*/
  char*     tag;           /* tag for the output file                       */
  char*     nBytesName;    /* FrSerData name for recording the frame size   */
  FrFile*   oFile;         /* output file pointer                           */
  FdAction* action;        /* this is the action structure                  */
};

void FdIFileProcess(FdIFile* f, FrameH* frame);
void FdOFileProcess(FdOFile* f, FrameH* frame);

/*--------------------------------------------------------------------------*/
int FdIFileNew(void *arg,   CfgDataType_t *data)
/*--------------------------------------------------------------------------*/
{                                        
  FdIFile *f;
  FdIO *fdIO;

  f = (FdIFile*) calloc(1,sizeof(FdIFile));
  if(f == NULL) CfgMsgAddFatal("malloc FdIFile failed");

  f->action = FdActionNew((FdAction**) arg, 
			  (void*) f, FdIFileProcess, "FDIN_FILE");

  if(f->action != *((FdAction**) arg)) 
    CfgMsgAddFatal("FDIN_FILE must be the first input key and not %s",
                   (*((FdAction**) arg))->type);

  f->action->userInfo = malloc(64);
  f->action->serData  = malloc(64);
  sprintf(f->action->userInfo, " ");
  sprintf(f->action->serData,  " ");

  /*------------------------------------------------------------ fill info---*/
  FrStrCpy(&(f->name), CfgParseGetNextString(data));
  f->start  =          CfgParseGetNextDec(data);
  f->length =          CfgParseGetNextDec(data);

  CfgMsgAddInfo("Open input file:%s start GPS=%d for %d seconds",
		f->name, f->start, f->length);

  /*------------------------------------get current GPS time if requested---*/
  if(f->start < 0) {
    f->start = FrGetCurrentGPS();
    CfgMsgAddInfo(" Set start GPS time to now:%d",f->start);}

  /*------------------------------------------------------open input file---*/
  f->iFile = FrFileINew(f->name);
  if(f->iFile == NULL) CfgMsgAddFatal("Cannot open input file");
  f->iFile->compress = FdIOGetCompressIn();

  /*--------------------------------get the time of the next frame to read---*/
  if(f->start > 0) {
    FrFileISetTime(f->iFile, f->start);
    f->requestedTime = f->start;}
  else  {
    f->requestedTime = FrFileITStart(f->iFile);}

  /*-------------------------------------get the end reading time if asked---*/
  if(f->length > 0) {f->endTime = f->requestedTime + f->length;}
  else              {f->endTime = 2e9;}

  /*---reduce the input Cm timeout to avoid waist time when reading a file---*/
  fdIO = FdIOGetRoot();
  fdIO->getFrTimeOut = 1.e-8;

  return(CFG_OK);
}
/*---------------------------------------------------------------------------*/
void FdIFileProcess(FdIFile* f, FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FdAction *next;
  double nBytes;

  /*-------------------if only a limited time was requested, are we done?---*/
  if((f->requestedTime >= f->endTime) || 
     (f->requestedTime < 0)) {
    CfgTerminate();
    return;}

  /*-----------------------no tag requested: regular sequential file read---*/
  nBytes = f->iFile->nBytes;
      
  if(f->action->tag == NULL) {
    frame = FrameRead(f->iFile);
    if(frame != NULL) f->requestedTime = frame->GTimeS + 1.e-9*frame->GTimeN 
      + 1.001*frame->dt;}

  /*------------------------------------------------------read with a tag---*/
  else {
    frame = FrameReadTChnl(f->iFile, f->requestedTime+.001, f->action->tag);
    if(frame != NULL) {
      frame->event = FrEventReadTF(f->iFile, f->action->tag, 
                                   frame->GTimeS + 1.e-9*frame->GTimeN, 
                                   frame->dt, 1, 0);}

    f->requestedTime = FrFileITNextFrame(f->iFile, f->requestedTime+0.001);}

  /*---------------------------------------------------------record info---*/
  nBytes = f->iFile->nBytes - nBytes;
  sprintf(f->action->serData,"units MB nBytesIn %g units counts", 
          nBytes/1024./1024.);
  if(nBytes < 1024*128) {
    sprintf(f->action->userInfo,"nBytesIn %.2f kB", nBytes/1024);}
  else { 
    sprintf(f->action->userInfo, "nBytes %.2f MB", nBytes/1024/1024);}

  /*---------------------terminte the job if we are at the end of the file---*/
  if (frame == NULL) CfgTerminate();

  /*------------------------------------------------------call next action---*/
  next = f->action->next;
  if(next != NULL) next->action(next->data, frame);

  return;
}
/*--------------------------------------------------------------------------*/
int FdOFileNew(void *arg,   CfgDataType_t *data)
/*--------------------------------------------------------------------------*/
{                                               /* Create a new output file */
  FdOFile *f;
  char *prefix, *fProbe, *p, *b, *dir, *end;
  FILE *fp;
  struct stat st;
 
  f = (FdOFile*) calloc(1,sizeof(FdOFile));
  if(f == NULL) CfgMsgAddFatal("malloc FdOFile failed");

  f->action = FdActionNew((FdAction**) arg, 
			  (void*) f, FdOFileProcess, "OFile");
  f->action->userInfo = malloc(64);
  f->action->serData  = malloc(64);
  f->action->userInfo[0]='\0';
  f->action->serData [0]='\0';

  FrStrCpy(&(f->name), CfgParseGetNextString(data));
  f->fLength        =  CfgParseGetNextDec(data);
  FrStrCpy(&(f->tag),  CfgParseGetNextString(data));
  f->dirPeriod      =  CfgParseGetNextDec(data);

  CfgMsgAddInfo("Open output file:%s-GPS with %d seconds per file; tag=%s",
		f->name, f->fLength, f->tag);

  /*-----------------------------------------------------------------------*/
  prefix = f->name;
  while(strstr(prefix,"/") != NULL) {prefix = strstr(prefix,"/")+1;}
  if(strlen(prefix) == 0) 
    CfgMsgAddFatal("Could not extract a prefix for the file name");
  f->nBytesName = malloc(20 + strlen(prefix));
  sprintf(f->nBytesName, "nBytes%s", prefix);
  CfgMsgAddInfo("  file prefix: %s  Record output frame size as %s", 
                prefix, f->nBytesName);

  /*-----------------------------------------------------------------------*/
  if(f->dirPeriod == 0)  {
    f->oFile = FrFileONewM(f->name, -1, CfgGetCmName(), f->fLength);}

  else { 
    CfgMsgAddInfo(" new directory will be created every %d seconds",
		  f->dirPeriod);
    if(f->dirPeriod < f->fLength) 
      CfgMsgAddFatal("Requested directory spam smaller than file length");
    f->oFile = FrFileONewMD(f->name, -1, CfgGetCmName(), f->fLength,
			    prefix, f->dirPeriod);}

  if(f->oFile == NULL) CfgMsgAddFatal("Cannot open output file");

  /*-------------------for files in /dev/shm: create the folder if needed---*/
  if(strncmp(f->name,"/dev/shm/",9) == 0) {
    FrStrCpy(&dir, f->name);
    end = dir + strlen(dir);
    p = dir+10;
    while(p < end) {
      b = strchr(p,'/');
      if(b == NULL) break;
      b[0] = '\0';
      if(stat(dir, &st) != 0) {
	CfgMsgAddInfo("Create folder %s",dir);
	if(mkdir(dir,0777) != 0) CfgMsgAddFatal(" cannot create folder");}
      b[0] = '/';
      p = b+1;}}

  /*-----------------------------------------------probe the write access---*/
  fProbe = malloc(strlen(f->name) + 20);
  sprintf(fProbe, "%s_Fd.probe", f->name);
  fp = fopen(fProbe,"w");
  if(fp == NULL) 
    CfgMsgAddFatal("Could not write output file(s): %s",strerror(errno));
  fclose(fp);
  remove(fProbe);
  free(fProbe);

  return(CFG_OK);
}
/*--------------------------------------------------------------------------*/
int FdOFileObsolete(void *arg,   CfgDataType_t *data)
/*--------------------------------------------------------------------------*/
{                                             
  CfgMsgAddFatal("The key FDOUT_SHORT_AND_LONG_FILES is not any more "
		 "available Use the FDOUT_FILE keys");

  return(CFG_OK);
}
/*---------------------------------------------------------------------------*/
void FdOFileProcessOne(FdOFile *f,
		       FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  double size;

  if(f->tag != NULL) FrameTag(frame,f->tag);

  size = f->oFile->nBytes;

  FrameWrite(frame, f->oFile);

  if(f->oFile->error != FR_OK) { 
    CfgMsgAddError("%d: frame write error on %s..", frame->GTimeS, f->name);
    f->action->state = CfgServerError;}
  else {
    f->action->state = CfgServerGolden;}

  if(f->tag != NULL) FrameUntag(frame);

  /*----be careful: when a new file is started, nBytes is reset to zero..---*/
  if(f->oFile->nFrames > 1) size = f->oFile->nBytes - size;
  else                      size = f->oFile->nBytes; 
  size = size/1024;
  if(size < 100) 
       sprintf(f->action->userInfo,"Write %.2f kB",size);
  else sprintf(f->action->userInfo,"Write %.2f MB",size/1024);

  if(size < 100) 
       sprintf(f->action->serData,
               "units kB %s %.2f units counts", f->nBytesName, size);
  else sprintf(f->action->serData,
               "units MB %s %.2f units counts", f->nBytesName, size/1024);

  return;
}
/*---------------------------------------------------------------------------*/
void FdOFileProcess(FdOFile *f,
		    FrameH* frame)
/*---------------------------------------------------------------------------*/
{
  FdAction *next;
  long nBytes;
  char *fName;

  if(frame != NULL) FdOFileProcessOne(f, frame);

  /*--------------------------frame == NULL means we are closing the files---*/
  else if(f->oFile == NULL) {
    CfgMsgAddInfo("No frame written on output file %s",f->name);}

  else if(f->oFile->frfd != NULL) {
    FrStrCpy(&fName, f->oFile->current->fileName);
    nBytes = FrFileOEnd(f->oFile);
    if(nBytes == 0) 
      CfgMsgAddError("Closing file error on %s ", fName);
    free(fName);
    f->oFile = NULL;}

  /*-----------------------------------------------------call next actions---*/
  next = f->action->next;
  if(next != NULL) next->action(next->data, frame);

  return;
}
/*---------------------------------------------------------------------------*/
void FdFileParserAdd(FdIO* fdIO)
/*---------------------------------------------------------------------------*/
{
 
  CfgParseGetFunctionAdd(fdIO->parser,"FDIN_FILE",
			 FdIFileNew, (void *) &(fdIO->actionsIn),
			 3, CfgString, CfgDec,  CfgDec);

  CfgParseGetFunctionAdd(fdIO->parser,"FDOUT_FILE",
			 FdOFileNew, (void *) &(fdIO->actionsOut),
			 4, CfgString, CfgDec,  CfgString,  CfgDec);
  CfgParseGetFunctionAdd(fdIO->parser,"FDOUT_SHORT_AND_LONG_FILES", 
			 FdOFileObsolete, NULL, 
                         7, CfgString, CfgString, CfgString, CfgDec, CfgDec,
                         CfgString, CfgDec);

  return;
}
