/*----------------------------------------------------------------------------*/
#include "FdIO.h"

typedef struct FdRangeGating FdRangeGating;

struct FdRangeGating {
  char*   hoftChannel;     /* name of the h(t) channel                        */
  double  mass;            /* mass used for the range computation             */
  int     nChunk;          /* number of FFT per input frame                   */
  double  minRange;        /* minimal range to apply gating                   */
  int     newHoftChannel;  /* if yes, create a new h(t) channel               */
  double  freqMin;         /* starting frequency for the horizon computation  */
  double  range;           /* last computed range                             */
  FrameH* previousFrame;
  FRBOOL  gatingPrevPart2; 
  FRBOOL  gatingCurrPart1;
  FRBOOL  gatingCurrPart2;
  FdAction* action;       /* this is the action process                       */
};

void FdRangeGatingProcess(FdRangeGating *rg, FrameH* frame);

/*----------------------------------------------------------------------------*/
double FdRGGetRange(
   FrVect *hoft, /* pointer to PSD vector                                     */
   double m1,    /* mass of source first object                               */
   double nuMin, /* lower frequency bound for SNR integration                 */
   double psdMin,
int nChunk)
/*----------------------------------------------------------------------------*/
{
  int i;
  double df, K, G, c, pi, Mpc, Msun, norm, freq, h2, snr;
  double eta, xLSO, nuLSO, invPSD, m2;
  FrvRFFT *fft;
  FrVect *psd;

  fft = FrvRFFTNewT("HNAO",hoft->dx[0]*hoft->nData/nChunk, 1.e9);    
  if(fft == NULL) CfgMsgAddFatal("FdHProcess: Could not create FFT");

  /*-------------------------------------------------------- build the PSD ---*/
  FrvRFFTFor(fft, hoft);
  psd = fft->amplitudeA;
  FrvMult(psd, psd, psd, NULL);
       
  pi = acos(-1.);
  c = 299792458.;
  G = 6.6726e-11;
  Mpc = 3.0856775807e22;
  Msun = 1.98892e30;

  /*--------------------------------------------- stop at last stable orbit---*/
  m2 = m1;
  eta = m1*m2/(m1+m2)/(m1+m2);
  xLSO = (2.*(-9. - eta + sqrt(1539. - 1008.*eta + 19.*eta*eta)))/
         (3.*(81. - 57.*eta + eta*eta));
  nuLSO = 1./(pi*(m1+m2)*4.92554e-6)*pow(xLSO,3./2.);

  norm = 5./4.*pi/6.*pow(G,5./3.)/c/c/c*pow(Msun,5./3.)/Mpc/Mpc*pow(pi,-7./3.);
  K = m1*m2/(m1+m2)*pow(m1+m2,2./3.);
  norm = norm*K;

  if(nuLSO > 1024) nuLSO =1024;

  df = psd->dx[0];
  snr = 0;
  for(i=1; i<psd->nData; i++) {
    freq = i*df;
   if(freq < nuMin) continue;
   if(freq > nuLSO) break;
   h2 = norm*pow(freq,-7./3.);
   if(psd->dataD[i] < psdMin) invPSD = 1./psdMin;
   else                       invPSD = 1./psd->dataD[i];
   snr += h2 * invPSD;}

  snr = sqrt(snr * 4. * df);
  FrvRFFTFree(fft);
  
  return(snr/8./2.26);
}
/*----------------------------------------------------------------------------*/
int FdRangeGatingNew(void *arg,   CfgDataType_t *data)
/*----------------------------------------------------------------------------*/
{
  FdRangeGating *rg;

  rg = (FdRangeGating*) calloc(1,sizeof(FdRangeGating));
  if(rg == NULL) CfgMsgAddFatal("malloc FdRangeGating failed");

  rg->action = FdActionNew((FdAction**) arg, (void*) rg, 
                           FdRangeGatingProcess, "RangeGating");

  FrStrCpy(&(rg->hoftChannel), CfgParseGetNextString(data));
  rg->mass           = CfgParseGetNextReal(data);
  rg->nChunk         = CfgParseGetNextDec(data);
  rg->minRange       = CfgParseGetNextReal(data);
  rg->newHoftChannel = CfgParseGetNextDec(data);
  
  if(rg->minRange > 0) {
    CfgMsgAddInfo("Range gating will be applied on %s, mass=%g nChunk=%d "
		  "minRange=%g",
 		  rg->hoftChannel, rg->mass, rg->nChunk, rg->minRange);
    if(rg->newHoftChannel == 0) 
         CfgMsgAddInfo(" h(t) channel will be updated");
    else CfgMsgAddInfo(" %s_Gated channel will be created",rg->hoftChannel);}
  else {
    CfgMsgAddInfo("Inspiral range will be computed on %s mass=%g nChunk=%d",
                  rg->hoftChannel, rg->mass, rg->nChunk);}

  rg->freqMin = 10;
  rg->previousFrame = NULL;
  rg->gatingCurrPart1 = FR_YES;
  rg->gatingCurrPart1 = FR_NO;

  rg->action->userInfo = malloc(100);
  sprintf(rg->action->userInfo," ");

  return(CFG_OK);
}
/*----------------------------------------------------------------------------*/
FrameH* FdRangeGatingProcessOne(FdRangeGating *rg, FrameH *frame)
/*----------------------------------------------------------------------------*/
/*
This fonction compute the inspiral range on one frame long.
If the range is bellow a threshold, a gating is apply on a frame length 
plus two windows of half a frame each.
This is why the code is working with two frames, in order to take care of the
glitches at the overlap and apply gating with a window.
------------------------------------------------------------------------------*/
{  
  double b, rangeC;
  char name[256];
  int i, expectedGTimeS, nData;
  FrVect *hoft, *hoftPrev, *hoftCurr, *hoftGated, *gate, *gateF;
  FRBOOL gatingPrevPart1, gatingPrevPart2, gatingCurrPart1, gatingCurrPart2;
  FrameH *previousFrame;
  FrProcData *proc, *pHoftPrev, *pHoftCurr;

  previousFrame = rg->previousFrame;
  rg->previousFrame = frame;

  if(frame == NULL) return(previousFrame);

  /*-----------------------------------------------------------startup case---*/
  if(previousFrame == NULL) {
    rg->gatingCurrPart1 = FR_YES;
    rg->gatingCurrPart2 = FR_YES;
    return(previousFrame);}

  gatingPrevPart1 = rg->gatingCurrPart1;
  gatingPrevPart2 = rg->gatingCurrPart2;
  gatingCurrPart1 = FR_NO;
  gatingCurrPart2 = FR_NO;
  
  /*------------------------------do nothing in case of frame discontinuity---*/
  expectedGTimeS = previousFrame->GTimeS + (int) previousFrame->dt;
  if(expectedGTimeS != frame->GTimeS) {
    return(previousFrame);}

  /*-----------------------------------get h(t) channel over the two frames---*/
  pHoftPrev = FrProcDataFind(previousFrame, rg->hoftChannel);
  pHoftCurr = FrProcDataFind(        frame, rg->hoftChannel);
  if(pHoftPrev == NULL || pHoftCurr == NULL) {
    rg->gatingPrevPart2 = FR_YES;
    return(previousFrame);}

  hoftPrev = pHoftPrev->data;
  hoftCurr = pHoftCurr->data;
  hoft = FrVectCopy(hoftPrev);
  if(hoft == NULL || hoftCurr == NULL) {
    rg->gatingPrevPart2 = FR_YES;
    return(previousFrame);}

  hoft->next = FrVectCopy(hoftCurr);
  nData = hoft->nData;
  hoft = FrVectConcat(hoft, hoft->GTime, 2 * nData * hoft->dx[0]);
  
  /*--------------compute and check the range on h(t) across the two frames---*/
  FrVectZoomInI(hoft, nData/2, nData); 
  rangeC = FdRGGetRange(hoft, rg->mass, rg->freqMin, 0, rg->nChunk);
  if(rangeC < rg->minRange) {
      gatingPrevPart2 = FR_YES;
      gatingCurrPart1 = FR_YES;}

  /*------------------------------- record the range for the previous frame---*/
 sprintf(name,"%s_Range",rg->hoftChannel);
 proc = FrProcDataNew(previousFrame, name, 2./frame->dt, 2, -32);
 proc->data->dataF[0] = rg->range;
 proc->data->dataF[1] = rangeC;
 FrVectSetUnitY(proc->data,"Mpc");

  /*------------------compute and check the range on h(t) for the new frame---*/
  FrVectZoomOut(hoft);
  FrVectZoomInI(hoft, nData, nData); 
  rg->range = FdRGGetRange(hoft, rg->mass, rg->freqMin, 0, rg->nChunk);
  if(rg->range < rg->minRange) {
      gatingCurrPart1 = FR_YES;
      gatingCurrPart2 = FR_YES;}
  FrVectFree(hoft);

  sprintf(rg->action->userInfo,"range: %.1f Mpc",rg->range);

  /*--------------------------------------returns if no gating is requested---*/
  if(rg->minRange <= 0) return(previousFrame);

  /*-------------------------------------------------record monitoring info---*/    
 sprintf(name,"%s_GateFlag",rg->hoftChannel);
 proc = FrProcDataNew(previousFrame, name, 2./frame->dt, 2, -32);
 proc->data->dataF[0] = gatingPrevPart1;
 proc->data->dataF[1] = gatingPrevPart2;
 FrVectSetUnitY(proc->data,"flag");

 /*--------------------computing gating on the first half of previous frame---*/
 sprintf(name,"%s_Gate",rg->hoftChannel);
 gate = FrVectCopyToD(hoftPrev, 0, name);
 if(gate == NULL) CfgMsgAddFatal("Could not create gate channel");
 FrVectSetUnitY(gate,"weight");

 b = 2*FRTWOPI/gate->nData;
 if(gatingPrevPart1 == FR_YES || 
   (gatingPrevPart2 == FR_YES && rg->gatingPrevPart2 == FR_YES)) {
     for(i = 0; i<gate->nData/2; i++) {gate->dataD[i] = 0;}}
 else if(rg->gatingPrevPart2 == FR_YES) {
     for(i = 0; i<gate->nData/2; i++) {gate->dataD[i] = 0.5*(1-cos(b*(i*.5)));}}     
 else if(gatingPrevPart2 == FR_YES) {
     for(i = 0; i<gate->nData/2; i++) {gate->dataD[i] = 0.5*(1+cos(b*(i*.5)));}}
 else {
     for(i = 0; i<gate->nData/2; i++) {gate->dataD[i] = 1;}}
     
 /*-------------------computing gating on the second half of previous frame---*/
 if(gatingPrevPart2 == FR_YES ||
   (gatingCurrPart1 == FR_YES && gatingPrevPart1 == FR_YES)) {
     for(i = gate->nData/2; i<gate->nData; i++) {gate->dataD[i] = 0;}}
 else if(gatingPrevPart1 == FR_YES) {
     for(i = gate->nData/2; i<gate->nData; i++) {gate->dataD[i] = 0.5*(1+cos(b*(i*.5)));}}     
 else if(gatingCurrPart1 == FR_YES) {
     for(i = gate->nData/2; i<gate->nData; i++) {gate->dataD[i] = 0.5*(1-cos(b*(i*.5)));}}
 else {
     for(i = gate->nData/2; i<gate->nData; i++) {gate->dataD[i] = 1;}}

 /*---------------convert gate data to float if needed; add it to the frame---*/
 if(hoftPrev->type == FR_VECT_4R) {
   gateF = FrVectCopyToF(gate, 1., NULL);
   FrVectFree(gate);
   gate = gateF;}

 FrProcDataNewV(previousFrame, gate);
 /*-----------------------------------------------------------apply gating ---*/
 hoftGated = FrVectCopy(hoftPrev);
 FrvMult(hoftGated, gate, hoftGated, NULL);

 if(rg->newHoftChannel != 0) {
   sprintf(name,"%s_Gated",rg->hoftChannel);
   FrVectSetName(hoftGated, name);
   FrProcDataNewV(previousFrame, hoftGated);}
 else {
   FrVectFree(pHoftPrev->data);
   pHoftPrev->data = hoftGated;}

 rg->gatingPrevPart2 = gatingPrevPart2;
 rg->gatingCurrPart1 = gatingCurrPart1;
 rg->gatingCurrPart2 = gatingCurrPart2;
 
 return(previousFrame);
}
/*----------------------------------------------------------------------------*/
void FdRangeGatingProcess(FdRangeGating *rg, FrameH* frame)
/*----------------------------------------------------------------------------*/
{
  FdAction *next;

  frame = FdRangeGatingProcessOne(rg, frame);

  next = rg->action->next;
  if(next != NULL) next->action(next->data, frame);

  return;
}
/*----------------------------------------------------------------------------*/
void FdRangeGatingAdd(FdIO* fdIO)
/*----------------------------------------------------------------------------*/
{
  CfgParseGetFunctionAdd(fdIO->parser, "FDIN_RANGE_GATING",
			 FdRangeGatingNew, (void *) &(fdIO->actionsIn), 5,
			 CfgString, CfgReal, CfgDec, CfgReal, CfgDec);

 return;
}
